! Example of custom trap table.
! This file is based on 'sparc-elf-4.4.2/src/libgloss/sparc_leon/locore_mvt.S'

! Command to compile project with custom trap table:
! sparc-elf-gcc -nostartfiles -O2 -mcpu=v8 -o name.out _trap_table.S _hardreset_init.S main.c

! 'start' is necessary entry point for linker (It is usially start of some area 0x00000000, 0x40000000 or ...).
! The traptable has to be the first code in boot (or must have 4096 bytes alignment).

! processor 1906VM016:
#define GPIO_ADDRESS          0x80100400
#define GPIO_DATA_IN_OFFSET         0x00
#define GPIO_DATA_OUT_OFFSET        0x04
#define GPIO_DIRECTION_OFFSET       0x08
#define GPIO_IRQ_MASK_OFFSET        0x0C

#define TRAP_JUMP(H)  sethi %hi(H), %l4; jmp %l4 + %lo(H); rd %psr, %l0; nop;

! Unexcpected trap will halt the processor by forcing it to error state.
#define BAD_TRAP ta 0; nop; nop; nop;
#define  x4_BAD_TRAP     BAD_TRAP;     BAD_TRAP;     BAD_TRAP;     BAD_TRAP;
#define x16_BAD_TRAP  x4_BAD_TRAP;  x4_BAD_TRAP;  x4_BAD_TRAP;  x4_BAD_TRAP;
#define x64_BAD_TRAP x16_BAD_TRAP; x16_BAD_TRAP; x16_BAD_TRAP; x16_BAD_TRAP;

! LEON 4 exceptions.
!
! Name:                           TrapType:
! reset                           0x00
! instruction_access_error        0x01 ("main" falt-tolerance trap. Uncorrectable error on READ instruction)
! illegal_instruction             0x02
! privileged_instruction          0x03
! fp_disabled                     0x04
! window_overflow                 0x05
! window_underflow                0x06
! mem_address_not_aligned         0x07
! fp_exception                    0x08
! data_access_exception           0x09 ("main" falt-tolerance trap. Uncorrectable error on READ data)
! tag_overflow                    0x0A
! watchpoint_detected             0x0B
! interrupt_level_1..15           0x11 - 0x1F
! cp_disabled                     0x24
! cp_exception                    0x28
! divide_exception                0x2A
! data_store_error                0x2B ("ancillary" falt-tolerance trap. Uncorrectable error on READ-modify-write)
! trap_instruction                0x80 - 0xFF

  .text
  .global start
  .global _trap_table
  ! '_hardreset_init' must exist in another assembler file of project (or can be placed in this file).
  .global _hardreset_init
start:
_trap_table:
  !TRAP_JUMP(_hardreset_init);             ! 00 reset trap
  flush;                                   ! 00 reset trap >>
  sethi %hi(_hardreset_init), %l4;
  jmp %l4 + %lo(_hardreset_init);
  nop;                                     ! 00 reset trap <<

  TRAP_JUMP(_ft_error_handler);            ! 01 instruction_access_exception
  BAD_TRAP;                                ! 02 illegal_instruction
  BAD_TRAP;                                ! 03 priveleged_instruction
  BAD_TRAP;                                ! 04 fp_disabled
  TRAP_JUMP(_window_overflow);             ! 05 window_overflow
  TRAP_JUMP(_window_underflow);            ! 06 window_underflow
  BAD_TRAP;                                ! 07 memory_address_not_aligned
  BAD_TRAP;                                ! 08 fp_exception
  TRAP_JUMP(_ft_error_handler);            ! 09 data_access_exception
  BAD_TRAP;                                ! 0A tag_overflow
  BAD_TRAP;                                ! 0B watchpoint_detected
  x4_BAD_TRAP;                             ! 0C - 0F undefined

  ! Interrupt entries.
  ! There are no interrupts handlers implemented. If they are need than see
  ! locore.S, irqtrap.S, irqrtap_fast.S  at 'sparc-elf-4.4.2/src/libgloss/sparc_leon' folder.
  x16_BAD_TRAP;                            ! 10 undefined, 11 - 1F interrupt level 1 - 15

  x4_BAD_TRAP;                             ! 20 - 23 undefined
  BAD_TRAP;                                ! 24 cp_disabled
  BAD_TRAP;                                ! 25 undefined
  BAD_TRAP;                                ! 26 undefined
  BAD_TRAP;                                ! 27 undefined
  BAD_TRAP;                                ! 28 cp_exception
  BAD_TRAP;                                ! 29 - undefined
  BAD_TRAP;                                ! 2A - divide_exception
  TRAP_JUMP(_ft_error_handler);            ! 2B - data_store_error (only at read-modify-write)
  x4_BAD_TRAP;                             ! 2C - 2F undefined
  x16_BAD_TRAP;                            ! 30 - 3F undefined
  x64_BAD_TRAP;                            ! 40 - 7F undefined

  ! Software traps.
  x64_BAD_TRAP; x64_BAD_TRAP;              ! 80 - FF
! End of _prom_trap_table.



! Window overflow / underflow handlers is based on the content of prominit.S from LEON mkprom >>
  .text
  .align  4
  .global _window_overflow
_window_overflow:
  mov  %wim, %l3                  ! Calculate next WIM
  mov  %g1, %l7
  srl  %l3, 1, %g1
  mov  %asr17, %l4                ! leon3 has nwindows in %asr17
  and  %l4, 0x1f, %l4
  sll  %l3, %l4 , %l4
  or   %l4, %g1, %g1

  save                            ! Get into window to be saved.
  mov  %g1, %wim
  nop; nop; nop
  std  %l0, [%sp + 0]
  std  %l2, [%sp + 8]
  std  %l4, [%sp + 16]
  std  %l6, [%sp + 24]
  std  %i0, [%sp + 32]
  std  %i2, [%sp + 40]
  std  %i4, [%sp + 48]
  std  %i6, [%sp + 56]
  restore                         ! Go back to trap window.
  mov  %l7, %g1
  jmp   %l1                       ! Re-execute save.
   rett %l2

  .global _window_underflow
  ! Window underflow trap handler on restore.
_window_underflow:
  mov  %wim, %l3                  ! Calculate next WIM
  sll  %l3, 1, %l4
  mov  %asr17, %l5                ! leon3 has nwindows in %asr17
  and  %l5, 0x1f, %l5
  srl  %l3, %l5, %l5
  or   %l5, %l4, %l5
  mov  %l5, %wim
  nop; nop; nop

  restore                         ! Two restores to get into the
  restore                         ! window to restore
  ldd  [%sp + 0], %l0             ! Restore window from the stack
  ldd  [%sp + 8], %l2
  ldd  [%sp + 16], %l4
  ldd  [%sp + 24], %l6
  ldd  [%sp + 32], %i0
  ldd  [%sp + 40], %i2
  ldd  [%sp + 48], %i4
  ldd  [%sp + 56], %i6
  save                            ! Get back to the trap window.
  save
  jmp   %l1                       ! Re-execute restore.
   rett %l2
! Window overflow / underflow handlers is based on the content of prominit.S from LEON mkprom <<


  .text
  .global _ft_error_handler
_ft_error_handler:
  ! Display trap type value on GPIO.
  rd   %tbr, %l6
  srl  %l6, 4, %l6
  and  %l6, 0xFF, %l6
  set  GPIO_ADDRESS, %l7
  st   %l6, [%l7 + GPIO_DATA_OUT_OFFSET]
  orn  %g0, %g0, %l6
  ! Set all pins to output.
  st   %l6, [%l7 + GPIO_DIRECTION_OFFSET]
  ! Enter to processor error mode.
  ta   0
